//=======================================================================================
// Project     : Charge-Arrival-Time Profiles
// Description : Basic type definitions, global database connection
// ======================================================================================

#ifndef CAT_PROFILES_BASE_H
#define CAT_PROFILES_BASE_H

#include <algorithm>
#include <limits>
#include <ostream>
//#include "cst.h"

namespace CATs {

using Time   = long long;    /// seconds
using Charge = double;       /// Whatt-hours (can also be used for energy/work in general, can be negative)
using CStID  = int;          /// Charge station ID
using NodeID = int;          /// Node IDs (as used by the router)

constexpr Time timeMinLimit = std::numeric_limits<Time>::min();

struct chfp{
    Time    t;
    Charge  ch;
};

using CHF = std::vector<chfp>;
using allCHF = std::unordered_map<CStID, CHF>;

using GraphRoute = std::vector<std::string>;    /// A path as a sequence of edge names

/*!
 * Interval [b, e)
*/
template<typename T>
struct Interval
{
    T b {std::numeric_limits<T>::max()};
    T e {std::numeric_limits<T>::min()};

    //! Check if interval is empty
    //! \return 1 or 0 if interval is empty
    bool empty() const { return e < b; };

    //!
    //! \return interval length
    T len() const { return e - b; };

    Interval operator-(T m) const { return {b - m, e - m}; };

    Interval operator+(T m) const { return {b + m, e + m}; };

    bool     operator==(const Interval& i) const { return b == i.b && e == i.e; };

    T midPoint() const { return b + (e - b) / 2; };  // avoids unnecessary overflow

    Interval& mbiExpand(const Interval& i)    //! Minimum bounding interval of this and i
    {
        b = std::min(b, i.b);
        e = std::max(e, i.e);
        return *this;
    };

    friend std::ostream& operator<< (std::ostream& out, const Interval& i)
    { return out << '{' << i.b << ',' << i.e << '}'; };
};

using WT = std::unordered_map<Time, Interval<Time>>;
using allWT = std::unordered_map<CStID, WT>;

using MapKatch = std::unordered_map<CStID, NodeID>;

using GraphRoute = std::vector<std::string>;
//using CStSet = std::vector<CATs::CSt>;

struct TE      /// Travel time, Energy used, and number of traversed edges
    {
        CATs::Time    time;
        CATs::Charge  energy;
        std::size_t   size;
        std::vector<Time>  edgeTimes;
    };

struct TEBounds   /// lower and upper bounds
    {
        TE first;
        TE last;
    };

struct TEBoundsPlus: TEBounds    ///  Augmented with two midpoints between first and last
    {
        TE mid1;
        TE mid2;
    };

struct StartEnd{
    CStID s;
    CStID e;
    };

struct Trip{
    long id;
    NodeID source;
    NodeID dest;
    Time t;
    };
}



#endif //CAT_PROFILES_BASE_H
