//=======================================================================================
// Project     : Charge-Arrival-Time Profiles
// Description : CAT profile
// ======================================================================================

#include "cat.h"
#include <cassert>

using namespace CATs;
using namespace std;

namespace CATs {
//-----------------------------------------------------------------------------
ostream &operator<<(ostream& out, const CAT& cat)
{
    out << '[';
    for (const auto& m: cat.els)
        out << '{' << m.ti << ',' << m.ci << "}, ";
    return out << "\b\b]" << endl;
}
}

//-----------------------------------------------------------------------------
CAT::CAT(const EVPath& evp, Interval<Time> ts, Charge cs)
{
    vector<CAT> tsCats;

    firstSlot = numeric_limits<int>::max();
    lastSlot  = numeric_limits<int>::min();

    // Iterating through all departure time slots
    auto tb = ts.b;
    auto te = (tb/Settings::delta + 1) * Settings::delta;
    for (auto last = false; !last; tb = te, te += Settings::delta)
    {
        if (te >= ts.e)
        {
            te = ts.e;
            last = true;
        }
        // [tb, te) a part of ts, intersecting with one time slot
        tsCats.push_back(build1Slot(evp, {tb, te}, cs));
        if (tsCats.back().firstSlot < firstSlot) firstSlot = tsCats.back().firstSlot;
        if (tsCats.back().lastSlot > lastSlot) lastSlot = tsCats.back().lastSlot;
    }

    if (tsCats.size() == 1) *this = move(tsCats.back());  // optimization for a single departure time slot
    else
    {
        // Iterating through all arrival time slots to construct the "sum" of all tsCats
        for (auto slot = firstSlot; slot <= lastSlot; slot++)
        {
            els.emplace_back();           // creates a default-initialized element
            for (const auto& e: tsCats)
                if (e.hasSlot(slot))
                {
                    els.back().ti.mbiExpand(e[slot].ti);
                    els.back().ci.mbiExpand(e[slot].ci);
                }
            feasible = feasible && els.back().ci.b >= 0;
            mbr.ci.mbiExpand(els.back().ci);  // compute charge MBI for the whole CAT
            assert(!els.back().ti.empty());      // Not possible for continuous travel-time functions!
        }
        mbr.ti = {els.front().ti.b, els.back().ti.e};   // Arrival time interval for the whole CAT
    }
}

//-----------------------------------------------------------------------------
CAT CAT::build1Slot(const EVPath& evp, Interval<Time> ts, Charge cs)
{
    CAT  rp;

    rp.mbr.ti    = evp.at(ts);   // arrival time interval corresponding to the departure interval ts
    rp.firstSlot = rp.mbr.ti.b / Settings::delta;
    rp.lastSlot  = rp.firstSlot - 1;

    // Iterating through all arrival time slots
    auto tb  = rp.mbr.ti.b;
    auto te  = (tb/Settings::delta + 1) * Settings::delta;
    for (auto last = false; !last; tb = te, te += Settings::delta)
    {
        rp.lastSlot++;
        if (te >= rp.mbr.ti.e)
        {
            te   = rp.mbr.ti.e;
            last = true;
        }
        // [tb, te) a part of ati, intersecting with one time slot
        rp.els.push_back({{tb,te}, evp.chl(cs, ts, evp.tt(ts, {tb, te}))});
        rp.feasible = rp.feasible && rp.els.back().ci.b >= 0;
        rp.mbr.ci.mbiExpand(rp.els.back().ci);       // compute charge MBI
    }
    return rp;
}

//-----------------------------------------------------------------------------
// TODO: Expand this
//
bool CAT::operator<(const CAT& other) const
{
    return mbr.ti.midPoint() < other.mbr.ti.midPoint();
}