//=======================================================================================
// Project     : Charge-Arrival-Time Profiles
// Description : CAT profile
// ======================================================================================

#ifndef CAT_PROFILES_CAT_H
#define CAT_PROFILES_CAT_H

#include "traffic_ev.h"
#include <vector>

namespace CATs {


//! Charge-Arrival-Time Profile class

class CAT
{
public:
    CAT() : firstSlot{-1}, lastSlot{-1} {} /*!< An empty CAT profile */

    //! Constructs a CAT for evp, when starting during ts, with a charge cs
    /*!
     \param evp Path of the vehicle
     \param ts interval of time
     \param cs charge level
    */
    CAT(const EVPath& evp, Interval<Time> ts, Charge cs);

    struct Element
    {
        Interval<Time>   ti;     /// Arrival time
        Interval<Charge> ci;     /// Charge left in a battery
    };

    bool isFeasible()      const { return feasible; };
    bool hasSlot(int slot) const { return firstSlot <= slot && slot <= lastSlot; };

    //! Easy access to a specific time-slot. Precondition: hasSlot(slot)
    /*!
     *
     * @param slot
     * @return
     */
    const Element& operator[] (int slot) const { return els[slot-firstSlot]; };

    bool operator<  (const CAT& other) const;  //! Comparison of two CATs

    /*!
     *
     * @param out
     * @param cat
     * @return
     */
    friend std::ostream& operator<< (std::ostream& out, const CAT& cat);

private:
    std::vector<Element>  els;
    Element               mbr;        /*!< minimum bounding rectangle of the CAT*/
    int                   firstSlot;  /*!< first element is in [firstSlot*delta, (firstSlot+1)*delta) */
    int                   lastSlot;   /*!< last element is in [lastSlot*delta, (lastSlot+1)*delta) */
    bool                  feasible{true};

    /*!  Constructs a CAT with the precondition that ts is contained in a single time slot
     *
     * @param evp
     * @param ts
     * @param cs
     * @return
     */
    static CAT build1Slot(const EVPath& evp, Interval<Time> ts, Charge cs);
};

/*!
 *
 * @param out
 * @param cat
 * @return
 */
std::ostream& operator<< (std::ostream& out, const CAT& cat);

}

#endif //CAT_PROFILES_CAT_H
