//=======================================================================================
// Project     : Charge-Arrival-Time Profiles
// Description : 1) Functions that call database functions to fetch/compute EV data
//               2) Integration with KaTCH router
// ======================================================================================

#ifndef CAT_PROFILES_INTEGRATION_H
#define CAT_PROFILES_INTEGRATION_H

#include "base.h"
#include "traffic_ev.h"

namespace CATs {

//-----------------------------------------------------------------------------
struct TE      /// Travel time, Energy used, number of traversed edges, and times it took to traverse the edges
{
    Time        time;
    Charge      energy;
    std::size_t size;
    std::vector<Time>  edgeTimes;    /// This is optional and can be empty!
};

struct TEBounds   /// lower and upper bounds
{
    TE first;
    TE last;
};

struct TEBoundsPlus: TEBounds    ///  Augmented with two midpoints between first and last
{
    TE mid1;
    TE mid2;
};

//-----------------------------------------------------------------------------
/// Computes/retrieves travel time/energy consumption interval (and subintervals if factor1 !=0.0 and factor2 != 1.0)
/// when leaving at a given time point
///
/// \post  In the return, TEBoundsPlus::edgeTimes are empty!
///
/// \param route
/// \param ts Start time in seconds
/// \param factor1 subinterval bounds as fractions of the returned interval
/// \param factor2
/// \param max_energy stops computing if more energy than this is used
TEBoundsPlus energy_consumption_tp(const GraphRoute& route, Time ts,
                                   Charge max_energy = std::numeric_limits<Charge>::max(),
                                   double factor1 = 0.0, double factor2 = 1.0);

//-----------------------------------------------------------------------------
/// Computes/retrieves travel time/energy consumption interval when leaving during a given time interval
///
/// \post  In the return, TEBoundsPlus::edgeTimes are filled
///
/// \param route
/// \param ts Start time interval in seconds
TEBounds     energy_consumption_ti(const GraphRoute& route, Interval<Time> ts);

//-----------------------------------------------------------------------------
CStSet getStationsAroundPath(const GraphRoute& route, int bufSize);

//-----------------------------------------------------------------------------
/// Calls the CaTCH router and returns the edge list of the computed route
///
/// \param source
/// \param destination
/// \param ts Start time in seconds
GraphRoute KaTCHRouting(NodeID source, NodeID destination, Time ts);

}

#endif //CAT_PROFILES_INTEGRATION_H
