//
// Created by linb on 2020-11-12.
//


#include "katch.h"

using std::string;

/*!Load KaTCH graph.
 *
 * @param btch_input_file_name parameter of the binary graph file
 * @return 0 if successfuly completed
 *
 * The method loads to global pointer graph data structure from file btch_input_file_name.
 */
int CATs::KaTCH::load(const string & btch_input_file_name) {
    std::vector<EdgeInfo> edge_list;
    edge_list = katch::btch_format::read_edges<EdgeInfo>(btch_input_file_name);
    if( edge_list.size() == 0 ){
        KATCH_ERROR("btch file: " << btch_input_file_name << " is empty");
        return EXIT_FAILURE;
    }
    katch::TchEaQuery::Graph KaTCHgraph{std::move(edge_list)};   /// Loaded graph from file as edgelist
    KATCH_STATUS("Graph base: " << KaTCHgraph.get_n_nodes() << " nodes, " << KaTCHgraph.get_n_edges() << " edges.\n");
    query_graph = std::make_unique<katch::TchEaQuery>(std::move(KaTCHgraph)); // Query graph is stored with unique pointer

    isLoaded = 1;
    return EXIT_SUCCESS;
}

/*!  Do query on KaTCH graph
 *
 * @param original_path  - address of the path where to store computed path
 * @param node_start   - node where to start from
 * @param node_end  - node to destination. Most of the time node_end is equal to path parameter destination
 * @param dep  - departure time in deci seconds  x/10 to get seconds
 * @param arr  - arrival time in deci second   x/10 to get seconds
 * @param rank - rank (we use 4).
 * @return 1 - always return true if does not break the code
 */

int CATs::KaTCH::query(katch::TchEaQuery::Path & original_path,
                       const uint32_t node_start,
                       const uint32_t node_end,
                       const double dep,
                       const double arr,
                       const uint32_t rank) {
    ///     initialize Demand structure
    //if( isLoaded == 1) return EXIT_FAILURE;
    std::cout<<"@";
    katch::demands_format::Demand demand{ node_start, node_end, dep, arr, rank };
    /// initilize query with the graph

    double max_abs_error = 0.0;
    double max_rel_error = 0.0;
    double total_abs_error = 0.0;
    double total_rel_error = 0.0;

    double max_abs_error_expand = 0.0;
    double max_rel_error_expand = 0.0;
    double total_abs_error_expand = 0.0;
    double total_rel_error_expand = 0.0;

    double total_route_extraction_and_expansion_time = 0.0;

    auto t1 = katch::util::time_stamp();
    /// double  - obtain travel time from start to destination
    const auto t_arr = query_graph->one_to_one(demand._start, demand._destination, demand._t_dep);
    /// compute absolute and relative errors of obtained travel time
    const double abs_error = fabs(t_arr - demand._t_arr);
    max_abs_error = std::max(max_abs_error, abs_error);
    total_abs_error += abs_error;

    const double rel_error = abs_error / fabs(demand._t_arr - demand._t_dep);
    max_rel_error = std::max(max_rel_error, rel_error);
    total_rel_error += rel_error;

    auto t_expand_begin = katch::util::time_stamp();
    /// obtain up-down-path from start to destination (requires katch::TchEaQuery::one_to_one called first)
    katch::TchEaQuery::Path up_down_path = query_graph->get_up_down_path(demand._destination);

    /// expand up-down-path to obtain path in the original road network
    original_path = query_graph->expand_path(up_down_path);

    /*    katch::EdgeIterator e;
    katch::NodeIterator u, v;
     KATCH_STATUS("Path_nodes. size:" << original_path.get_n_edges() << " \n");
        KATCH_STATUS("(");

        for( size_t  i = 0 ; i < original_path.get_n_edges(); ++i ){
          u = std::get<0>(original_path.get_edge(i));
          e = std::get<1>(original_path.get_edge(i));
          v = std::get<2>(original_path.get_edge(i));
          KATCH_CONTINUE_STATUS("\"" << u <<"-" << v << "\",");

        }
        KATCH_CONTINUE_STATUS(")\n");
*/

    auto t_expand_end = katch::util::time_stamp();
    total_route_extraction_and_expansion_time += katch::util::get_duration_in_seconds(t_expand_begin, t_expand_end);
    /// compute absolute and relative errors of the travel time of the expanded path
    const double t_arr_expand = original_path.get_t_arr(original_path.get_n_edges());

    const double abs_error_expand = fabs(t_arr_expand - demand._t_arr);
    max_abs_error_expand = std::max(max_abs_error_expand, abs_error_expand);
    total_abs_error_expand += abs_error_expand;

    const double rel_error_expand = abs_error_expand / demand._t_arr;
    max_rel_error_expand = std::max(max_rel_error_expand, rel_error_expand);
    total_rel_error_expand += rel_error_expand;
    return EXIT_SUCCESS;
}
