//=======================================================================================
// Project     : Charge-Arrival-Time Profiles
// Description : Traffic and ev related information: Path; tt, wt functions
// ======================================================================================

#include "traffic_ev.h"
//#include "integration.h"
#include <cmath>
#include <cassert>

using namespace CATs;
using namespace std;
using namespace settings;

namespace
{
inline Time xtion(Time at, Interval<Time> ts, Time bttx, Time ettx, bool rup)
{
    double dttx = ettx - bttx;
    auto   rez  = (dttx*(at - ts.b) + ts.len()*static_cast<double>(bttx)) / (ts.len() + dttx);

    return rup ? ceil(rez) : floor(rez);
}
}

//-----------------------------------------------------------------------------
Interval<Time> EVPath::tt(Interval<Time> ts, Interval<Time> ta) const
{
    Interval<Time>  ri;
    auto            btti = tt(ts.b);
    auto            etti = tt(ts.e);

    auto            btte = ta.e - ts.b;   // tt.e corresponding to ta.e at ts.b
    auto            ette = ta.e - ts.e;   // tt.e corresponding to ta.e at ts.e
    auto            ettb = ta.b - ts.e;   // tt.b corresponding to ta.b at ts.e
    auto            bttb = ta.b - ts.b;   // tt.b corresponding to ta.b at ts.b

    // Computing the beginning of the interval
    if (etti.b <= ettb)
        ri.b = ettb;
    else if (ettb < etti.b <= ette)
        if (btti.b <= bttb)
            ri.b = min(etti.b, xtion(ta.b, ts, btti.b, etti.b, false));
        else  // btti.b > bttb
            ri.b = min(btti.b, etti.b);
    else  // etti.b > ette
        if (btti.b >= bttb)
            ri.b = min(btti.b, xtion(ta.e, ts, btti.b, etti.b, false));
        else  // btti.b < bttb
            ri.b = xtion(ta.b, ts, btti.b, etti.b, false);

    // Computing the end of the interval
    if (btti.e >= btte)
        ri.e = btte;
    else if (bttb <= btti.e < btte)
        if (etti.e >= ette)
            ri.e = max(btti.e, xtion(ta.e, ts, btti.e, etti.e, true));
        else  // etti.e < ette
            ri.e = max(btti.e, etti.e);
    else  // bttb > btti.e
        if (etti.e <= ette)
            ri.e = max(etti.e, xtion(ta.b, ts, btti.e, etti.e, true));
        else  // etti.e > ette
            ri.e = xtion(ta.e, ts, btti.e, etti.e, true);

    return ri;
}

//-----------------------------------------------------------------------------
Interval<Time> EVPath::tt(Time ts) const
{
    if (ts % Settings::delta)   // in the middle of a time slot: perform interpolation
    {
        auto tsl = ts - ts % Settings::delta;
        auto ttl = loadIfNeeded(tsl).tti;
        auto ttr = loadIfNeeded(tsl + Settings::delta).tti;

        // interpolate between ttl and ttr
        return {static_cast<Time>(floor(ttl.b + (ttr.b - ttl.b) *
                                            static_cast<double>(ts % Settings::delta) / Settings::delta)),
                static_cast<Time>(ceil(ttl.e + (ttr.e - ttl.e) *
                                            static_cast<double>(ts % Settings::delta) / Settings::delta))};
    }
    return loadIfNeeded(ts).tti;
}

//-----------------------------------------------------------------------------
EVPath::ttec& EVPath::loadIfNeeded(Time ts) const
{
    auto it = cash.find(ts);

    if (it == cash.end())
    {
        auto ec_data = energy_consumption_tp(path, ts);

        // The smaller tt will usually result in more energy used (due to larger speed)
        // That's why a swap of energy interval ends is usually necessary
        //
        if (ec_data.first.energy > ec_data.last.energy)
            swap(ec_data.first.energy, ec_data.last.energy);
        it = cash.emplace(ts, ttec{{ec_data.first.time, ec_data.last.time},
                                   {ec_data.first.energy, ec_data.last.energy}}).first;
    }
    return it->second;
}

//-----------------------------------------------------------------------------
Interval<Charge> EVPath::chl(Charge cs, Interval<Time> ts, Interval<Time> tt) const
{
    // It is possible that sometimes smaller tt will result in less energy used => more left in a battery
    // i.e., that's why the ends of charge interval could be reversed (but returned from here correctly)
    //
    auto  cei = [&tt, &cs, this](Time dt)  // Lambda to compute the energy interval for dt departure time
                {
                    Interval<Charge> ret;

                    auto tti     = loadIfNeeded(dt).tti;
                    auto ttl     = tti.len();
                    auto factor1 = static_cast<double>(tt.b - tti.b) / ttl;
                    if (factor1 < 0.0) factor1 = 0.0;
                    auto factor2 = static_cast<double>(tt.e - tti.b) / ttl;
                    if (factor2 > 1.0) factor2 = 1.0;
                    if (factor1 != 0.0 || factor2 != 1.0)
                    {
                        auto ec_data = energy_consumption_tp(path, dt, cs, factor1, factor2);
                        if (ec_data.mid1.size < path.size()) // not enough energy
                            ec_data.mid1.energy = cs + 1;    // will be negative, when subtracted from cs...
                        if (ec_data.mid2.size < path.size())
                            ec_data.mid2.energy = cs + 1;    //... to signify infeasibility
                        ret = {ec_data.mid1.energy, ec_data.mid2.energy};
                        if (ret.b > ret.e) swap(ret.b, ret.e);
                    } else
                        ret = cash[dt].eci;    // note: interval ends can be more than cs (infeasible)

                    return ret;
                };
    // Compute the energy intervals for ts.b and ts.e departure times
    //
    auto chtsb = cei(ts.b);
    auto chtse = cei(ts.e);

    return {cs - max(chtsb.e, chtse.e), cs - min(chtsb.b, chtse.b)};
    // return {testECFunction(tt.b), testECFunction(tt.e)};     // Just for testing
}

//------------------------------------------------------------------------------
namespace {
// finds nid as the first id of an edge in a path and returns its index,
// else returns one-past-end
// Assumes a path is a sequence of edges encoded as "nodeid-nodeid" strings
auto findNodeID (const GraphRoute& path, NodeID nid)
{
    auto   snid = to_string(nid);
    return find_if (cbegin(path), cend(path),
                    [&snid](const auto& edge)
                    { return !edge.compare(0, edge.find('-'), snid); }) - cbegin(path);
}
}

//-----------------------------------------------------------------------------
CStSet EVPath::CStsAlong(Charge maxEnergy, int bufDist, Interval<Time> ts)
{
    if (path.empty()) return CStSet{};

    auto fullDrive   = energy_consumption_ti(path, ts);
    auto energyDrive = energy_consumption_tp(path, ts.midPoint(), maxEnergy);

    path.resize(min(energyDrive.first.size, energyDrive.last.size)); // truncate the path

    auto ret = getStationsAroundPath(path, bufDist);

    // Compute timeToD for all the result CSts:
    // First, compute suffix sums of fullDrive.first.edgeTimes
    auto& timeToDs = fullDrive.first.edgeTimes;
    assert(!timeToDs.empty());  // This has to be true for a non-empty path (see the first line)
    timeToDs.push_back(0);      // To simplify the following two loops
    for (auto ri = timeToDs.size() - 1; ri; ri--)
        timeToDs[ri-1] += timeToDs[ri];

    // Then, assign them to close-by CSts (based on CSt::NodeID)
    for (auto& st: ret)
        st.timeToD = timeToDs[findNodeID(path, st.getNodeID())];

    return ret;
}


