//=======================================================================================
// Project     : Charge-Arrival-Time Profiles
// Description : Traffic and ev related information: Path; tt, wt functions
// ======================================================================================

#ifndef CAT_PROFILES_TRAFFIC_EV_H
#define CAT_PROFILES_TRAFFIC_EV_H

#include "cst.h"
#include <unordered_map>
#include <string>
#include "dbwrapper.h"
#include "triprouter.h"

namespace CATs {

/*!------------------------------------------------------------------------------------------
 *  An expected traversal of a specific path by a specific ev
 *  Used in two scenarios:
 *  1) Constructed as a "guide route" to destination and charging stations around it are found
 *  2) Constructed as a leg of a final route and its CAT profile is computed
 *
 * \pre All interval ts parameters are assumed to be contained in a single time slot!
 */

class EVPath
{
public:
    struct ttec
    {
        Interval<Time>    tti;   /// travel-time interval
        Interval<Charge>  eci;   /// energy consumed interval
    };

    /// Maps departure time to travel-time/energy-consumed rectangle
    using TT = std::unordered_map<Time, ttec>;

             EVPath() {};
    explicit EVPath(GraphRoute p) : path{std::move(p)} {};
    explicit EVPath(TT&& csh) : cash{std::move(csh)}       // For testing
             { path.resize(1); };

    /// Check if EVPath is empty
    /// \return bool
    bool empty() const { return path.empty(); };

    /// Arrival-time interval, when leaving during ts
    /// \pre Travel time function respects the FIFO property
    ///
    /// \param ts
    /// \return
    Interval<Time>   at(Interval<Time> ts) const { return {at(ts.b).b, at(ts.e).e}; };

    /// Travel-time and arrival-time intervals when leaving at ts
    ///
    /// \param ts
    /// \return
    Interval<Time>   tt(Time ts) const;
    Interval<Time>   at(Time ts) const { return tt(ts) + ts; };

    /// Travel-time interval when leaving during ts and arriving during ta
    ///
    /// \param ts
    /// \param ta
    /// \return
    Interval<Time>   tt(Interval<Time> ts, Interval<Time> ta) const;

    /// Charge left in a battery, if starting charge is cs, leaving some time during ts, travel-time in the interval tt
    ///
    /// \param cs
    /// \param ts
    /// \param tt
    /// \return
    Interval<Charge> chl(Charge cs, Interval<Time> ts, Interval<Time> tt) const;

    /// Get charging stations in a buffer along the route,
    /// but not further than maxEnergy away from the beginning, as driven if leaving in the middle of ts
    /// CSt::timeToDs are filled according to slowest driving when leaving at ts.b
    ///
    /// \note  This method uses the internal structure of path (more than a black-box id)
    ///
    /// \post this->path is cut according to maxEnergy
    ///
    /// \param maxEnergy  stop "driving" when reaching max-energy
    /// \param bufDist    the size of the buffer in Euclidean meters around the path
    /// \param ts         interval of departure times
    CStSet CStsAlong(Charge maxEnergy, int bufDist, Interval<Time> ts);

private:
    GraphRoute  path;

    CStSet getStationsAroundPath(const GraphRoute& route, int bufSize)
    { return TripRouter::getDB().getStationsAroundPath(route, bufSize); };

    TEBoundsPlus energy_consumption_tp(const GraphRoute& route, Time ts,
                                       Charge max_energy = std::numeric_limits<Charge>::max(),
                                       double factor1 = 0.0, double factor2 = 1.0) const
    { return TripRouter::getDB().energy_consumption_tp(route, ts, max_energy, factor1, factor2); };

    TEBounds energy_consumption_ti(const GraphRoute& route, Interval<Time> ts) const
    { return TripRouter::getDB().energy_consumption_ti(route, ts); };

    /// A cash of computed tt and ec intervals
    mutable TT cash;

    ttec& loadIfNeeded(Time ts) const;  /// returns an element in cash; puts it there if needed
};

}
#endif //CAT_PROFILES_TRAFFIC_EV_H

