/*
Name:
	snap_additional_points_on_network - Generates network with additional network nodes from specified sql
	
Description:
	sourcePointSql: Additional points to add to network. The sql must be in form "SELECT id::integer, the_geom::geometry FROM table"
	networkEdgesSql: Current network geometries. The sql must be in form "SELECT id::integer, the_geom::geometry, km::float8, kmh::integer FROM table"
	outputTableName: The new network table that will be created
Usage:
	SELECT snap_additional_points_on_network(
		'SELECT stop_id::integer AS id, the_geom AS the_geom FROM gtfs_stops WHERE stop_id IN( ''10415'',''10417'',''10424'')',
		'SELECT id_0::integer AS id, geom AS the_geom, km, kmh FROM roads_nz_wellington',
		'hutt_pax.tmp_test'
	);
 SELECT snap_additional_points_on_network(
		'SELECT id AS id, geom AS the_geom FROM de_chstations',
		'SELECT id AS id, geom_way AS the_geom, km, kmh FROM grm_2po_4pgr',
		'de_chstations_snap'
	);
Author:
	Adam Lawrence <alaw005@gmail.com>
*/
-- https://docs.pgrouting.org/2.0/en/doc/src/tutorial/topology.html#topology


DROP FUNCTION IF EXISTS snap_additional_points_on_network(text, text, text, text);
CREATE OR REPLACE FUNCTION snap_additional_points_on_network(
	sourcePointSql text DEFAULT 'SELECT id::bigint, the_geom:geometry FROM pointstable', 
	networkEdgesSql text DEFAULT 'SELECT id:bigint, the_geom:geometry, km::float, kmh::integer FROM networktable', 
	outputPointsTableName text DEFAULT 'tmp_new_nodes',
	outputTableName text DEFAULT 'my_generated_network'
)
	RETURNS text AS
$BODY$
DECLARE

BEGIN

	-- Create temp table to hold the new nodes snapped to the network
	RAISE NOTICE 'Creating outputPointsTableName table to hold new nodes';
	EXECUTE 'DROP TABLE IF EXISTS ' || outputPointsTableName;
	EXECUTE 'CREATE TABLE ' || outputPointsTableName || '  (
		id serial PRIMARY KEY,
		ch_station_id bigint,
		old_edge_id bigint,
		the_geom geometry
	)';
	EXECUTE 'CREATE INDEX idx_geom_' || outputPointsTableName || ' ON  ' || outputPointsTableName || ' USING gist(the_geom)'; 
	-- 4326),3785
	-- Locate source points to the nearest network edge and insert into temp table
	RAISE NOTICE 'Inserting new nodes into temporary table (this may take some time .. around 100 ms per node)';
	EXECUTE 'INSERT INTO ' || outputPointsTableName || ' (ch_station_id, old_edge_id, the_geom)
		SELECT 
			DISTINCT ON (a.id) 
			a.id AS ch_station_id,
			b.id AS old_edge_id,
			ST_ClosestPoint(b.the_geom, a.the_geom) AS the_geom
		FROM (' || sourcePointSql || ') AS a 
			JOIN (' || networkEdgesSql || ') AS b ON ST_DWithin(a.the_geom::GEOGRAPHY, b.the_geom::GEOGRAPHY, 100)
		ORDER BY  
			a.id, 
			ST_Distance(a.the_geom, b.the_geom)';
		
	RETURN 'OK';  

END;
$BODY$
  LANGUAGE plpgsql VOLATILE;
  