-- https://github.com/eclipse/sumo/blob/0531526c2eac1918508d779c64e321fa0d85566b/src/utils/emissions/HelpersEnergy.cpp
-- ms         - speed on ending the edge
-- last_ms    - speed on starting the edge 
-- time       - time to travel given edge
-- delta_h    - height difference along the edge (end point - start point)
-- edge_dist  - distance along the edge, length of edge in meters
-- edge_radius- rotational radius of the edge approximating edte as an arc
CREATE OR REPLACE FUNCTION EV_energy(ms float, last_ms float, time_s float, delta_h float, edge_dist float, edge_radius float) RETURNS float AS $$
DECLARE
  energy float := 0;
 
  mass float := 1000;
  internalmomentofinertia float :=  0.01;
  fronsurfacearea float := 5.0;  
  airdragcoef float := 0.6; 
  rolldragcoef float := 0.01;
  radialdragcoef float := 0.5;
  constantpowerintake float := 100;
  propulsioneficiency float := 0.9;
  recuperationeficiency float := 0.8;
  
BEGIN
    --calculate potential energy difference -- energyDiff = mass * 9.81 * sin(DEG2RAD(slope)) * SPEED2DIST(v);
    energy := energy + mass*9.81*(delta_h);
	-- kinetic energy difference of vehicle --energyDiff += 0.5 * mass * (v * v - lastV * lastV);
    energy := energy + 0.5 * mass * (ms * ms - last_ms * last_ms);
	-- add rotational energy diff of internal rotating elements energyDiff += INTERNALMOMENTOFINERTIA * (v * v - lastV * lastV);
    energy := energy + internalmomentofinertia * (ms * ms - last_ms * last_ms);
    --  Energy loss through Air resistance [Ws] -- energyDiff += 0.5 * 1.2041 * FRONTSURFACEAREA * AIRDRAGCOEFFICIENT * v * v * SPEED2DIST(v);
    energy := energy + 0.5 * 1.2041 * fronsurfacearea * airdragcoef * ms * ms * edge_dist;
    -- Energy loss through Roll resistance [Ws]   -- energyDiff += ROLLDRAGCOEFFICIENT * 9.81 * mass * SPEED2DIST(v);
    energy := energy + rolldragcoef * 9.81 * mass * edge_dist;
    -- Energy loss through friction by radial force [Ws] -- energyDiff += RADIALDRAGCOEFFICIENT * mass * v * v / radius;
	IF edge_radius IS NOT null THEN
	  energy := energy + radialdragcoef * mass * ms * ms / edge_radius;
	END IF;
    -- Energy loss through constant loads (e.g. A/C) [Ws]  -- energyDiff += CONSTANTPOWERINTAKE; //(per 1 sec!!)
    energy := energy + constantpowerintake * time_s;
    if energy > 0 THEN 
	    -- Assumption: Efficiency of myPropulsionEfficiency when accelerating	-- energyDiff /= PROPULSIONEFFICIENCY;
        energy := energy/propulsioneficiency;
    ELSE
        -- Assumption: Efficiency of myRecuperationEfficiency when recuperating -- energyDiff *= RECUPERATIONEFFICIENCY;
        energy := energy * recuperationeficiency;
		-- Insaf Sagaama, Amine Kchiche, Wassim Trojet, Farouk Kamoun Improving The Accuracy of The Energy Consumption Model for
	    --if (a != 0) {
        --       energyDiff *= (1 / exp(RECUPERATIONEFFICIENCY_BY_DECELERATION / fabs(a)));
        --   }
    END IF;
    -- convert from [Ws], i.e. [J]  to [Wh] (3600s / 1h):
    energy := energy / 3600;
  RETURN energy;
END;
$$ LANGUAGE plpgsql;


--DROP FUNCTION  energy_on_path(int[], int, float);
-- calculate energy on path. 
-- edges - path segment ids
-- start_time - travel start time
-- factor - speed scale factor default = 0 no scale
-- max_energy - the max battery energy. if this energy is less than path required energy
--              function will calculate only the subpath possible to go with max energy
-- RETURNS: travel time on calculated path, energy used on path, path segments 
CREATE OR REPLACE FUNCTION energy_on_path(edges int[], start_time int, max_energy float default null, path_speed float[] default null, path_time float[] default null) RETURNS 
 json  AS $$
DECLARE
  
  energy      float := 0; -- total energy used
  delta_energy float := 0; -- energy on last segment
  
  travel_time float := 0; -- total travel time
  last_ms     float := 0;
  ms          float := 0;
  delta_time  float := 0;
  delta_h     float := 0;
 
  new_path_speed  float[];
  new_path_time   float[];
  
  cnt         int;
  edge        pgrout_export%rowtype;
  path_size   int;
    
BEGIN
  IF path_speed is null THEN 
    path_size := array_length(edges, 1);
  ELSE
    path_size := array_length(path_speed, 1);
  END IF;
  cnt := 1;
  
  last_ms := 0; -- start from speed = 0 m/s 
  LOOP
    SELECT INTO edge * FROM pgrout_export WHERE id = edges[cnt];
    
	 RAISE NOTICE 'path_speed % path_size %', path_speed, path_size;
	 IF path_speed is null THEN -- use ttf speed profile if speed on path is not given
	  IF cnt = path_size THEN
	    ms := 0; -- stop till speed 0 for last edge
	  ELSE
	    ms := get_ttf_speed(edge.trfk_ms, edge.ms, start_time + travel_time::int);
	  END IF;
	  delta_time := get_ttf_time(edge.time, edge.trfk_time, start_time + travel_time::int) ;
      RAISE NOTICE ' speed path  is null';	  
    ELSE
	  RAISE NOTICE ' speed path is not null ';
	  ms := path_speed[cnt];
	  delta_time := path_time[cnt];
	END IF;
	
	delta_h    := edge.target_h - edge.source_h;
 
    delta_energy := EV_energy(ms, last_ms, delta_time, delta_h, edge.km*1000, edge.edge_radius);

    --RAISE NOTICE 'total aplied % delta time ttl % and factor delta %', travel_time, get_ttf_time(edge.time, edge.trfk_time, start_time + travel_time::int), delta_time;
    RAISE NOTICE 'cnt % time% e% delta e %', cnt, travel_time,  energy, delta_energy;
    EXIT WHEN max_energy is not null AND (energy + delta_energy) > max_energy;
    
	energy := energy + delta_energy; 	
	travel_time := travel_time + delta_time;
	last_ms := ms;
	
	new_path_speed[cnt] := ms;
	new_path_time[cnt] := delta_time;
	
	cnt := cnt + 1; -- now count points to the next segent
	EXIT WHEN cnt > path_size;
  END LOOP;
  
  IF cnt > path_size THEN -- remove last segment if no energy
	cnt:= cnt - 1;
  END IF;

  -- Build the JSON Response:
  RETURN json_build_object(
    'time', travel_time,
    'energy', energy,
    'size', cnt,
    'path_speed', new_path_speed,
    'path_time', new_path_time);
 
END;
$$ LANGUAGE plpgsql;



-- test
-- select * from energy_on_path('{32,33,44,55,66,77,88,99}'::INT[], 16*36000, 100 );


-- ttl_interval_bounds will calculate time slot bounds for given time (it will do mod 24 hour, so 25 will be converted to 24)

CREATE OR REPLACE FUNCTION ttl_interval_bounds(start_time int, int_size int default 36000) RETURNS int[] AS $$
DECLARE
  time_first  int := 0;
  time_last   int := 0;
  slot        int := 0;
BEGIN
    
  time_first := mod(864000 + int_size*floor(start_time/int_size)::int, 864000);
  time_last  := mod(int_size*(floor(start_time/int_size)::int + 1), 864000);
  slot       := floor(start_time/int_size)::int;
  
  RETURN array [time_first, time_last, slot];
END;
$$ LANGUAGE plpgsql;

-- test
--select ttl_interval_bounds(700000);



CREATE OR REPLACE FUNCTION energy_on_path_DT_interval(edges int[], start_time int, factor float , max_energy float default null) RETURNS json AS $$
DECLARE
  slot_start   int[];
  int_first    json;
  int_last     json;
  int_midle    json;
  path_speed_1 float[];
  path_speed_2 float[];
  path_time_1  float[];
  path_time_2  float[];
   
  path_speed float[];
  path_time  float[];
  
  path_size  int;
  cnt        int := 1;
  
  
  
BEGIN
  slot_start := ttl_interval_bounds(start_time, 36000);
  
  int_first := energy_on_path(edges, slot_start[1]);
  int_last  := energy_on_path(edges, slot_start[2]);
   
  path_speed_1 := TRANSLATE(int_first->>'path_speed', '[]','{}')::FLOAT[];
  path_speed_2 := TRANSLATE(int_last->>'path_speed', '[]','{}')::FLOAT[];
  
  path_time_1 := TRANSLATE(int_first->>'path_time', '[]','{}')::FLOAT[];
  path_time_2 := TRANSLATE(int_last->>'path_time', '[]','{}')::FLOAT[];
  
  path_size := LEAST((int_first->>'size')::INT ,  (int_last->>'size')::INT);
  cnt := 1;
 
  LOOP 
    IF path_time_1[cnt] = path_time_2[cnt] THEN
	  path_time[cnt] := path_time_1[cnt];
	  path_speed[cnt] := path_speed_1[cnt];
	ELSE
	  path_time[cnt] := abs(path_time_1[cnt] - path_time_2[cnt]) * factor + LEAST(path_time_1[cnt], path_time_2[cnt]);
      path_speed[cnt] := GREATEST(path_speed_1[cnt], path_speed_2[cnt]) - abs(path_speed_1[cnt] - path_speed_2[cnt]) * factor;	  
	END IF;
	
    EXIT WHEN cnt >= path_size;
	cnt := cnt + 1; -- now count points to the next segent
  END LOOP;
  
  int_midle := energy_on_path(edges, start_time, max_energy, path_speed, path_time);
  
  RETURN json_build_object(
      'first', int_first,
	  'mid',  int_midle,
      'last', int_last);
END;
$$ LANGUAGE plpgsql; 


--select * from energy_on_path_DT_interval('{32,33,44,55,66,77,88,99}'::INT[], 16*36000, 100 );

--select s->'first'->>'time',s->'mid'->>'time', s->'last'->>'time',
--       s->'first'->>'path_time',s->'mid'->>'path_time',s->'mid'->>'path_speed',s->'last'->>'path_time', s
--	   from energy_on_path_DT_interval('{32,33,44,55,66,77,88,99}'::INT[], 16*36000, 0.25, 100 ) as s;



CREATE OR REPLACE FUNCTION energy_on_path_interval(edges int[], start_time int, granularity int default 0 ) 
RETURNS  table (ttm float, energy float) AS $$
DECLARE
 
  time_first  int := 0;
  time_last   int := 0;

  int_first   float[];
  int_last    float[];
  int_tmp     float[];
  scale       float;
  
  int_size    int := 0;   

  cnt         int := 1;
  
BEGIN
 
  time_first := mod(864000 + 36000*floor((start_time - 36000)/36000)::int, 864000);
  time_last  := mod(36000*floor((start_time + 36000)/36000)::int, 864000);

  RAISE NOTICE 'time first % time last %', time_first, time_last;

  int_first := energy_on_path(edges, time_first, 0);
  int_last  := energy_on_path(edges, time_last, 0);
  
  RAISE NOTICE 'int first % int last %', int_first, int_last;
  --int first {828038,21.5886940811043} int last {36038,21.6986066266257}
  
  int_size  := abs(int_last[1]*10 - int_first[1]*10)/granularity;
  scale     := sign(int_last[1] - int_first[1])*granularity/(int_first[1]*10);
  
  RAISE NOTICE ' size % > scale % , granularity % ', int_size, scale, granularity;
  
  ttm := int_first[1]*10; -- KATCH uses 0.1 sec
  energy := int_first[2];
  return next;  

  LOOP
    EXIT WHEN cnt >= int_size - 1;
	int_tmp := energy_on_path(edges, time_first, scale*cnt);
	ttm := int_tmp[1]*10; -- KATCH uses 0.1 sec
	energy := int_tmp[2];
	return next;

   cnt := cnt + 1;
  END LOOP;
  
  ttm := int_last[1]*10; -- KATCH uses 0.1 sec
  energy := int_last[2];
  return next;

END;
$$ LANGUAGE plpgsql;

select * from energy_on_path_interval('{32,33,44,55,66,77,88,99}'::INT[], 16*36000, 1000 );

select * from energy_on_path('{32,33,44,55,66,77,88,99}'::INT[], 16*36000, 0, 100 );


select energy_on_path('{32,33}'::INT[], 36000*a), a from generate_series(0, 23) as a;

select energy_on_path('{32,33}'::INT[], 36000, 0.1*a), a from generate_series(0, 5) as a;

select energy_on_path('{32,33}'::INT[], 0, 0.1*a), a from generate_series(0, 5) as a;

select energy_on_path_interval('{32,33}'::INT[], 0, granularity int default  );






CREATE OR REPLACE FUNCTION energy_on_path_time_slot(edges int[], start_time int, interval_size int default 36000 ) 
RETURNS  table (ttm float, energy float) AS $$
DECLARE
 
  time_first  int := 0;
  time_last   int := 0;
  time_tmp    int := 0;

  slot_start  int[];
  slot_first  int[];
  slot_last   int[]; 
  
  int_first   float[];
  int_last    float[];
  int_tmp     float[];
  scale       float;
  
  direction   int; -- +1 positive, -1 negative
  
  int_size    int := 0;   

  cnt         int := 1;
  
BEGIN
  
  slot_start := ttl_interval_bounds(start_time, interval_size);

  RAISE NOTICE 'time first % time last %', slot_start[1], slot_start[2];

  int_first := energy_on_path(edges, slot_start[1], 0);
  int_last  := energy_on_path(edges, slot_start[2], 0);
  
  RAISE NOTICE 'int first % int last %', int_first, int_last;
  --int first {828038,21.5886940811043} int last {36038,21.6986066266257}
  
  slot_first := ttl_interval_bounds(slot_start[1] + int_first[1], interval_size);
  slot_last  := ttl_interval_bounds(slot_start[2] + int_last[1], interval_size);
  direction := sign(int_last[1] - int_first[1]);
  --int_size  := abs(int_last[1]*10 - int_first[1]*10)/interval_size;
  
  --RAISE NOTICE ' size % > scale % , granularity % ', int_size, scale, granularity;
  
  ttm := int_first[1]*10; -- KATCH uses 0.1 sec
  energy := int_first[2];
  return next;  

  time_tmp := slot_first[(1.5 + 0.5*direction)::int]; -- map 1 to 2;  -1 to 1
  
  LOOP
    EXIT WHEN ((direction > 0) AND time_tmp > slot_last[1]) OR 
	          ((direction < 0) AND time_tmp < slot_last[2]);
	
	scale     := (time_tmp - int_first[1]*10)/(int_first[1]*10);
	
	int_tmp := energy_on_path(edges, time_first, scale*cnt);
	ttm := int_tmp[1]*10; -- KATCH uses 0.1 sec
	energy := int_tmp[2];
	return next;

   cnt := cnt + 1;
  END LOOP;
  
  ttm := int_last[1]*10; -- KATCH uses 0.1 sec
  energy := int_last[2];
  return next;

END;
$$ LANGUAGE plpgsql;