-- https://github.com/eclipse/sumo/blob/0531526c2eac1918508d779c64e321fa0d85566b/src/utils/emissions/HelpersEnergy.cpp
-- ms         - speed on ending the edge
-- last_ms    - speed on starting the edge 
-- time       - time to travel given edge
-- delta_h    - height difference along the edge (end point - start point)
-- edge_dist  - distance along the edge, length of edge in meters
-- edge_radius- rotational radius of the edge approximating edte as an arc
CREATE OR REPLACE FUNCTION EV_energy(ms float, last_ms float, time_s float, delta_h float, edge_dist float, edge_radius float, car_physics json default null) RETURNS float AS $$
DECLARE
  energy float := 0;
 
  mass float := 1000;
  internalmomentofinertia float :=  0.01;
  fronsurfacearea float := 5.0;  
  airdragcoef float := 0.6; 
  rolldragcoef float := 0.01;
  radialdragcoef float := 0.5;
  constantpowerintake float := 100;
  propulsioneficiency float := 0.9;
  recuperationeficiency float := 0.8;
  
BEGIN
    IF car_physics is not null THEN
      mass := (car_physics->>'mass')::float;
      internalmomentofinertia := (car_physics->>'internalmomentofinertia')::float;
      fronsurfacearea  := (car_physics->>'fronsurfacearea')::float;  
      airdragcoef := (car_physics->>'airdragcoef')::float; 
      rolldragcoef := (car_physics->>'rolldragcoef')::float;
      radialdragcoef := (car_physics->>'radialdragcoef')::float;
      constantpowerintake := (car_physics->>'constantpowerintake')::float;
      propulsioneficiency := (car_physics->>'propulsioneficiency')::float;
      recuperationeficiency := (car_physics->>'recuperationeficiency')::float;
	END IF;
	
    --calculate potential energy difference -- energyDiff = mass * 9.81 * sin(DEG2RAD(slope)) * SPEED2DIST(v);
    energy := energy + mass*9.81*(delta_h);
	-- kinetic energy difference of vehicle --energyDiff += 0.5 * mass * (v * v - lastV * lastV);
    energy := energy + 0.5 * mass * (ms * ms - last_ms * last_ms);
	-- add rotational energy diff of internal rotating elements energyDiff += INTERNALMOMENTOFINERTIA * (v * v - lastV * lastV);
    energy := energy + internalmomentofinertia * (ms * ms - last_ms * last_ms);
    --  Energy loss through Air resistance [Ws] -- energyDiff += 0.5 * 1.2041 * FRONTSURFACEAREA * AIRDRAGCOEFFICIENT * v * v * SPEED2DIST(v);
    energy := energy + 0.5 * 1.2041 * fronsurfacearea * airdragcoef * ms * ms * edge_dist;
    -- Energy loss through Roll resistance [Ws]   -- energyDiff += ROLLDRAGCOEFFICIENT * 9.81 * mass * SPEED2DIST(v);
    energy := energy + rolldragcoef * 9.81 * mass * edge_dist;
    -- Energy loss through friction by radial force [Ws] -- energyDiff += RADIALDRAGCOEFFICIENT * mass * v * v / radius;
	IF edge_radius IS NOT null THEN
	  energy := energy + radialdragcoef * mass * ms * ms / edge_radius;
	END IF;
    -- Energy loss through constant loads (e.g. A/C) [Ws]  -- energyDiff += CONSTANTPOWERINTAKE; //(per 1 sec!!)
    energy := energy + constantpowerintake * time_s;
    if energy > 0 THEN 
	    -- Assumption: Efficiency of myPropulsionEfficiency when accelerating	-- energyDiff /= PROPULSIONEFFICIENCY;
        energy := energy/propulsioneficiency;
    ELSE
        -- Assumption: Efficiency of myRecuperationEfficiency when recuperating -- energyDiff *= RECUPERATIONEFFICIENCY;
        energy := energy * recuperationeficiency;
		-- Insaf Sagaama, Amine Kchiche, Wassim Trojet, Farouk Kamoun Improving The Accuracy of The Energy Consumption Model for
	    --if (a != 0) {
        --       energyDiff *= (1 / exp(RECUPERATIONEFFICIENCY_BY_DECELERATION / fabs(a)));
        --   }
    END IF;
    -- convert from [Ws], i.e. [J]  to [Wh] (3600s / 1h):
    energy := energy / 3600;
    if energy is null then 
	  RAISE NOTICE 'EV_energy(%, %, %, %, %, %)', ms, last_ms, time_s, delta_h, edge_dist, edge_radius;
	end if;
  RETURN energy;
END;
$$ LANGUAGE plpgsql;
--test
select EV_energy(30.0, 60.0, 100, 300, 400, 2000);

select EV_energy(30.0, 60.0, 100, 300, 400, 2000, json_build_object(
  'mass', 1500,
  'internalmomentofinertia', 0.01,
  'fronsurfacearea', 3.0,
  'airdragcoef', 0.3,
  'rolldragcoef', 0.01,
  'radialdragcoef', 0.3,
  'constantpowerintake', 100,
  'propulsioneficiency', 0.9,
  'recuperationeficiency', 0.8)
);


--DROP FUNCTION  energy_on_path(text[], int, float, float[], float[]);
-- calculate energy on path. 
-- edges - path segment ids
-- start_time - travel start time
-- max_energy - the max battery energy. if this energy is less than path required energy
--              function will calculate only the subpath possible to go with max energy
-- interval_size - defines the interval size in time units to determine the speed for min and max values
-- interval_bound - specifies min value (0) and max value for (1). 0 - slowest time 1 fastest 

-- RETURNS: travel time on calculated path, energy used on path, path segments 
DROP FUNCTION energy_on_path(text[], int, int, int, float, float, json, float[], float[]); 


CREATE OR REPLACE FUNCTION energy_on_path(edges text[], start_time int, interval_size int, interval_bound int, speedyfactor float, max_energy float default null, car_physics json default null, path_speed float[] default null, path_time float[] default null) RETURNS 
 json  AS $$
DECLARE
  ttf_price   float :=0; -- price from time profile function based on travel time
  
  energy      float := 0; -- total energy used
  delta_energy float := 0; -- energy on last segment
  
  travel_time float := 0; -- total travel time
  last_ms     float := 0;
  ms          float := 0;
  delta_time  float := 0;
  delta_h     float := 0;
 
  new_path_speed  float[];
  new_path_time   float[];
  
  cnt         int;
  edge        t_pgrout_export%rowtype;
  path_size   int;
    
BEGIN
  IF path_speed is null THEN 
    path_size := array_length(edges, 1);
  ELSE
    path_size := array_length(path_speed, 1);
  END IF;
  cnt := 1;
  
  -- RAISE NOTICE 'path_speed % path_size %', path_speed, path_size;
  
  last_ms := 0; -- start from speed = 0 m/s 
  LOOP
    SELECT INTO edge * FROM t_pgrout_export WHERE sourcetarget = edges[cnt];
    
	 --RAISE NOTICE 'path_speed % path_size %', path_speed, path_size;
	 IF path_speed is null THEN -- use ttf speed profile if speed on path is not given
	  
	  IF interval_size = 0 THEN
	    IF cnt = path_size THEN
	      ms := 0; -- stop till speed 0 for last edge
	    ELSE
	      ms := speedyfactor*get_ttf_speed(edge.trfk_ms, edge.ms, start_time + travel_time::int);
	    END IF;
	    delta_time := (1/speedyfactor)*get_ttf_time(edge.time, edge.trfk_time, start_time + travel_time::int) ;
	  ELSE 
	    ttf_price = get_ttf_price(start_time + travel_time::int, interval_size, interval_bound);
	    IF cnt = path_size THEN
	      ms := 0; -- stop till speed 0 for last edge
	    ELSE
	      ms := calculate_ttf_speed(edge.trfk_ms, edge.ms, ttf_price);
	    END IF;
	    delta_time := calculate_ttf_time(edge.time, edge.trfk_time, ttf_price) ;
      END IF;
	  
	  --RAISE NOTICE ' speed path is null';	  
    ELSE
	  --RAISE NOTICE ' speed path is not null ';
	  ms := path_speed[cnt];
	  delta_time := path_time[cnt];
	END IF;
	
	delta_h    := COALESCE( edge.target_h - edge.source_h, 0);
 
    delta_energy := EV_energy(ms, last_ms, delta_time, delta_h, edge.km*1000, edge.edge_radius, car_physics);

    --RAISE NOTICE 'cnt % time% e% delta e %', cnt, travel_time,  energy, delta_energy;
    EXIT WHEN max_energy is not null AND (energy + delta_energy) > max_energy;
    
	energy := energy + delta_energy; 	
	travel_time := travel_time + delta_time;
	last_ms := ms;
	
	new_path_speed[cnt] := ms;
	new_path_time[cnt] := delta_time;
	
	cnt := cnt + 1; -- now count points to the next segent
	EXIT WHEN cnt > path_size;
  END LOOP;
  -- remove last segment count 
  cnt:= cnt - 1;

  -- Build the JSON Response:
  RETURN json_build_object(
    'time', travel_time,
    'energy', energy,
    'size', cnt,
    'path_speed', new_path_speed,
    'path_time', new_path_time);
 
END;
$$ LANGUAGE plpgsql;



-- test
--  select * from energy_on_path('{"1-26731","26731-26735","26735-363051","363051-119619","119619-243307","243307-404182","404182-404183","404183-135287","135287-403967","403967-403968","403968-403980","403980-235853"}'::text[], 16*36000, 4*36000, -1 , 1);


-- convinience function will calculte energy by given path_id which is a id in table path_log.
CREATE OR REPLACE FUNCTION energy_on_path_id(path_id integer, start_time int, interval_size int, interval_bound int, speedyfactor float, max_energy float default null, car_physics json default null, path_speed float[] default null, path_time float[] default null) RETURNS 
 json  AS $$
DECLARE
   edges text[];
BEGIN
   select path into edges from path_log where path_log.id = path_id;
   return  energy_on_path(edges, start_time, interval_size, interval_bound, speedyfactor, max_energy, car_physics, path_speed, path_time);   
END; 
$$
LANGUAGE plpgsql;
-- test
--  select * from energy_on_path_id(1, 16*36000, 4*36000, -1 , 1);

-- convinience function will calculte energy by given path_name which is a name in table path_log.
CREATE OR REPLACE FUNCTION energy_on_path_nameid(path_name text, start_time int, interval_size int, interval_bound int, speedyfactor float, max_energy float default null, car_physics json default null, path_speed float[] default null, path_time float[] default null) RETURNS 
 json  AS $$
DECLARE
   edges text[];
BEGIN
   select path into edges from path_log where path_log.name = path_name;
   return  energy_on_path(edges, start_time, interval_size, interval_bound, speedyfactor, max_energy, car_physics,  path_speed, path_time);   
END; 
$$
LANGUAGE plpgsql;
-- test
--  select * from energy_on_path_nameid('1-6', 16*36000, 4*36000, -1 , 1);


DROP FUNCTION energy_on_path_minmax_bounds(edges text[], start_time_first int, start_time_last int, max_energy float, car_physics json);

CREATE OR REPLACE FUNCTION energy_on_path_minmax_bounds(edges text[], start_time_first int, start_time_last int,  max_energy float default null, car_physics json default null) RETURNS 
 json  AS $$
DECLARE
  ttf_price_min   float :=0; -- price from time profile function based on travel time
  ttf_price_max   float :=0; -- price from time profile function based on travel time
  
  energy_first      float := 0; -- total energy used
  energy_last       float := 0; -- total energy used
  
  delta_energy_first float := 0; -- energy on last segment
  delta_energy_last  float := 0; -- energy on last segment
  
  travel_time_first  float := 0; -- total travel lowest time
  travel_time_last  float := 0; -- total travel fastes time
  
  last_ms_first     float := 0;
  last_ms_last      float := 0;
  
  ms_first          float := 0;
  ms_last           float := 0;
  
  delta_time_first  float := 0;
  delta_time_last   float := 0;
  delta_h     float := 0;
 
  new_path_speed_first  float[];
  new_path_time_first   float[];
  new_path_speed_last  float[];
  new_path_time_last   float[];
  
  cnt         int;
  cnt_first   int;
  cnt_last    int;
  
  edge        t_pgrout_export%rowtype;
  path_size   int;
    
BEGIN
  path_size := array_length(edges, 1);
  cnt := 1;
  cnt_first := 0;
  cnt_last := 0;
  
  last_ms_first := 0; -- start from speed = 0 m/s 
  last_ms_last := 0; -- start from speed = 0 m/s 
 
 LOOP
    SELECT INTO edge * FROM t_pgrout_export WHERE sourcetarget = edges[cnt];
    
	ttf_price_max = get_ttf_min_max_interval_price(start_time_first + travel_time_first::int, start_time_last + travel_time_last::int, 1);
	ttf_price_min = get_ttf_min_max_interval_price(start_time_first + travel_time_first::int, start_time_last + travel_time_last::int, -1);
	
	RAISE NOTICE 'time slots from % - % price range % - %', start_time_first + travel_time_first::int, start_time_last + travel_time_last::int, ttf_price_min, ttf_price_max;
	  
	 -- IF cnt = path_size THEN
	      ms_first := 0; -- stop till speed 0 for last edge
	      ms_last  := 0;
	  --ELSE	  
	      ms_first := calculate_ttf_speed(edge.trfk_ms, edge.ms, ttf_price_min);
		  ms_last  := calculate_ttf_speed(edge.trfk_ms, edge.ms, ttf_price_max);	  
	  --END IF;
	  
	  delta_time_first := calculate_ttf_time(edge.time, edge.trfk_time, ttf_price_min);
      delta_time_last := calculate_ttf_time(edge.time, edge.trfk_time, ttf_price_max);
      
	  delta_h := COALESCE( edge.target_h - edge.source_h, 0);
      
      delta_energy_first := EV_energy(ms_first, last_ms_first, delta_time_first, delta_h, edge.km*1000, edge.edge_radius, car_physics);
      RAISE NOTICE 'energ first % delta %  cnt % cnt first % cnt last %', (energy_first + delta_energy_first), delta_energy_first,  cnt, cnt_first, cnt_last;
	
      IF max_energy is not null AND 
	    ((energy_first + delta_energy_first) > max_energy OR (cnt - 1) > cnt_first)
	  THEN
	   RAISE NOTICE '-- virsijo pirmas %', energy_first + delta_energy_first;
      ELSE 	  
	    energy_first := energy_first + delta_energy_first;
	    travel_time_first := travel_time_first + delta_time_first;
	    last_ms_first := ms_first;
	    cnt_first := cnt;
	    new_path_speed_first[cnt] := ms_first;
	    new_path_time_first[cnt] := delta_time_first;
      END IF; 

      delta_energy_last := EV_energy(ms_last, last_ms_last, delta_time_last, delta_h, edge.km*1000, edge.edge_radius, car_physics); 
	 RAISE NOTICE 'energ last %  delta % cnt % cnt first % cnt last %', (energy_last + delta_energy_last),delta_energy_last, cnt, cnt_first, cnt_last;
	
	  IF max_energy is not null AND 
	    ((energy_last + delta_energy_last) > max_energy OR (cnt - 1) > cnt_last)
	  THEN
	   RAISE NOTICE '-- virsijo antras % ', energy_last + delta_energy_last;
      ELSE 	  
	    energy_last := energy_last + delta_energy_last;
 	    travel_time_last := travel_time_last + delta_time_last;
	    last_ms_last := ms_last;
	    cnt_last := cnt;
	    new_path_speed_last[cnt] := ms_last;
	    new_path_time_last[cnt] := delta_time_last;
	  END IF;
	   RAISE NOTICE 'energ first % eneerg last %  cnt % cnt first % cnt last %', (energy_first), (energy_last), cnt, cnt_first, cnt_last;
	
	  EXIT WHEN max_energy is not null AND cnt > cnt_first and cnt > cnt_last;
	
	cnt := cnt + 1; -- now count points to the next segent
	EXIT WHEN cnt > path_size;
  END LOOP;
  
  IF cnt > path_size THEN -- remove last segment if no energy
	cnt:= cnt - 1;
  END IF;

  -- Build the JSON Response:
  RETURN json_build_object(
      'first', json_build_object(
				'time', travel_time_first,
                'energy', energy_first,
                'size', cnt_first,
                'path_speed', new_path_speed_first,
                'path_time', new_path_time_first),
	  'last',  json_build_object(
				'time', travel_time_last,
                'energy', energy_last,
                'size', cnt_last,
                'path_speed', new_path_speed_last,
                'path_time', new_path_time_last)
	   );
  
END;
$$ LANGUAGE plpgsql;

-- test
select s->'first'->>'time', s->'last'->>'time', s->'first'->>'size', 
        s->'first'->>'energy', s->'last'->>'energy', s->'last'->>'size'  ,s
from energy_on_path_minmax_bounds('{"1-26731","26731-26735","26735-363051","363051-119619","119619-243307","243307-404182","404182-404183","404183-135287","135287-403967","403967-403968","403968-403980","403980-235853"}'::text[], 16*36000, 17*36000) as s;

-- 
-- convinience function will calculate energy by given path_id which is a id in table path_log.
CREATE OR REPLACE FUNCTION energy_on_path_id_minmax_bounds(path_id integer, start_time_first int, start_time_last int,  max_energy float default null, car_physics json default null) RETURNS 
 json  AS $$
DECLARE
   edges text[];
BEGIN
   select path into edges from path_log where path_log.id = path_id;
   return  energy_on_path_minmax_bounds(edges, start_time_first, start_time_last,  max_energy, car_physics);   
END; 
$$
LANGUAGE plpgsql;
-- test
select s->'first'->>'time', s->'last'->>'time', s->'first'->>'size', 
        s->'first'->>'energy', s->'last'->>'energy', s->'last'->>'size'  ,s
from energy_on_path_id_minmax_bounds(1, 16*36000, 17*36000) as s;

-- 
-- convinience function will calculate energy by given path_name which is a name in table path_log.
CREATE OR REPLACE FUNCTION energy_on_path_nameid_minmax_bounds(path_name text, start_time_first int, start_time_last int,  max_energy float default null, car_physics json default null) RETURNS 
 json  AS $$
DECLARE
   edges text[];
BEGIN
   select path into edges from path_log where path_log.name = path_name;
   return  energy_on_path_minmax_bounds(edges, start_time_first, start_time_last,  max_energy, car_physics);   
END; 
$$
LANGUAGE plpgsql;
-- test
select s->'first'->>'time', s->'last'->>'time', s->'first'->>'size', 
        s->'first'->>'energy', s->'last'->>'energy', s->'last'->>'size'  ,s
from energy_on_path_nameid_minmax_bounds('1-6', 16*36000, 17*36000) as s;



CREATE OR REPLACE FUNCTION energy_on_path_interval_clipper(int_first json, int_last json, edges text[], start_time int, factor1 float, factor2 float, max_energy float default null, car_physics json default null) RETURNS json AS $$
DECLARE
  int_midle1    json;
  int_midle2    json;
  path_speed_1 float[];
  path_speed_2 float[];
  path_time_1  float[];
  path_time_2  float[];
   
  path_speed float[];
  path_time  float[];
  
  path_size  int;
  cnt        int := 1;
  
  
BEGIN
    
  path_speed_1 := TRANSLATE(int_first->>'path_speed', '[]','{}')::FLOAT[];
  path_speed_2 := TRANSLATE(int_last->>'path_speed', '[]','{}')::FLOAT[];
  
  path_time_1 := TRANSLATE(int_first->>'path_time', '[]','{}')::FLOAT[];
  path_time_2 := TRANSLATE(int_last->>'path_time', '[]','{}')::FLOAT[];
  
  path_size := LEAST((int_first->>'size')::INT ,  (int_last->>'size')::INT);

  IF factor1 = 0 THEN
    int_midle1 = int_first;
  ELSIF factor1 = 1 THEN
    int_midle1 = int_last;
  ELSE
    cnt := 1;
    LOOP 
      IF path_time_1[cnt] = path_time_2[cnt] THEN
	    path_time[cnt] := path_time_1[cnt];
	    path_speed[cnt] := path_speed_1[cnt];
	  ELSE
	    path_time[cnt] := (path_time_2[cnt] - path_time_1[cnt]) * factor1 + path_time_1[cnt];
        path_speed[cnt] := path_speed_1[cnt] - (path_speed_1[cnt] - path_speed_2[cnt]) * factor1;	  
	  END IF;
	
      EXIT WHEN cnt >= path_size;
	  cnt := cnt + 1; -- now count points to the next segent
    END LOOP;	
    -- max energy jau nenaudojam nes gi intervalai jau suskaiciouti
    int_midle1 := energy_on_path(edges, start_time, 0, 0, 0, null, car_physics, path_speed, path_time);
  END IF;
  
  IF factor2 = 0 THEN
    int_midle2 = int_first;
  ELSIF factor2 = 1 THEN
    int_midle2 = int_last;
  ELSE
    cnt := 1;
    LOOP 
      IF path_time_1[cnt] = path_time_2[cnt] THEN
	    path_time[cnt] := path_time_1[cnt];
	    path_speed[cnt] := path_speed_1[cnt];
	  ELSE
	    path_time[cnt] := (path_time_2[cnt] - path_time_1[cnt]) * factor2 + path_time_1[cnt];
        path_speed[cnt] := path_speed_1[cnt] - (path_speed_1[cnt] - path_speed_2[cnt]) * factor2;	  
	  END IF;
	
      EXIT WHEN cnt >= path_size;
	  cnt := cnt + 1; -- now count points to the next segent
    END LOOP;
  
    int_midle2 := energy_on_path(edges, start_time, 0,0, 0, null, car_physics, path_speed, path_time);
  END IF;
  
  RETURN json_build_object(
      'first', int_first,
	  'mid1',  int_midle1,
	  'mid2',  int_midle2,	  
      'last',  int_last);
END;
$$ LANGUAGE plpgsql; 


-- DT dynamic time interval - will use min value and max values from interval of given size

CREATE OR REPLACE FUNCTION energy_on_path_DT_interval(edges text[], start_time int, factor1 float, factor2 float, max_energy float default null, car_physics json default null) RETURNS json AS $$
DECLARE
  int_first    json;
  int_last     json;
  
  res          json;  
  
BEGIN
  
  int_first := energy_on_path(edges, start_time, 2*36000, -1, 1, max_energy, car_physics);
  int_last  := energy_on_path(edges, start_time, 2*36000, 1, 1, max_energy, car_physics);
  
  res := energy_on_path_interval_clipper(int_first, int_last, edges, start_time, factor1, factor2, max_energy, car_physics);
  
  RETURN res;
END;
$$ LANGUAGE plpgsql; 

-- select s->'first'->>'time',s->'mid1'->>'time',s->'mid2'->>'time', s->'last'->>'time', 
--        s->'first'->>'energy',s->'mid1'->>'energy', s->'mid2'->>'energy', s->'last'->>'energy',   
--        s->'first'->>'path_time',s->'mid'->>'path_time',s->'mid'->>'path_speed',s->'last'->>'path_time', s
--	   from energy_on_path_DT_interval('{"1-26731","26731-26735","26735-363051","363051-119619","119619-243307","243307-404182","404182-404183","404183-135287","135287-403967","403967-403968","403968-403980","403980-235853"}'::text[], 16*36000, 0.25, 0.75 ) as s;


--	  path_time[cnt] := abs(path_time_1[cnt] - path_time_2[cnt]) * factor1 + LEAST(path_time_1[cnt], path_time_2[cnt]);
--     path_speed[cnt] := GREATEST(path_speed_1[cnt], path_speed_2[cnt]) - abs(path_speed_1[cnt] - path_speed_2[cnt]) * factor1;	  

CREATE OR REPLACE FUNCTION energy_on_path_id_DT_interval(path_id integer, start_time int, factor1 float, factor2 float, max_energy float default null, car_physics json default null) RETURNS json AS $$
DECLARE
   edges text[];
BEGIN
   select path into edges from path_log where path_log.id = path_id;
   return  energy_on_path_DT_interval(edges, start_time, factor1, factor2, max_energy, car_physics);   
END; 
$$
LANGUAGE plpgsql;
-- test
 select s->'first'->>'time',s->'mid1'->>'time',s->'mid2'->>'time', s->'last'->>'time', 
        s->'first'->>'energy',s->'mid1'->>'energy', s->'mid2'->>'energy', s->'last'->>'energy',   
       s->'first'->>'path_time',s->'mid'->>'path_time',s->'mid'->>'path_speed',s->'last'->>'path_time', s
	   from energy_on_path_id_DT_interval(1, 16*36000, 0.25, 0.75 ) as s;
 select s->'first'->>'time',s->'mid1'->>'time',s->'mid2'->>'time', s->'last'->>'time', 
        s->'first'->>'energy',s->'mid1'->>'energy', s->'mid2'->>'energy', s->'last'->>'energy',   
       s->'first'->>'path_time',s->'mid'->>'path_time',s->'mid'->>'path_speed',s->'last'->>'path_time', s
	   from energy_on_path_id_DT_interval(1, 16*36000, 0.25, 0.75, 400000, json_build_object(
  'mass', 2500,
  'internalmomentofinertia', 0.01,
  'fronsurfacearea', 3.0,
  'airdragcoef', 0.3,
  'rolldragcoef', 0.01,
  'radialdragcoef', 0.3,
  'constantpowerintake', 100,
  'propulsioneficiency', 0.9,
  'recuperationeficiency', 0.8) ) as s;



CREATE OR REPLACE FUNCTION energy_on_path_nameid_DT_interval(path_name text, start_time int, factor1 float, factor2 float, max_energy float default null) RETURNS json AS $$
DECLARE
   edges text[];
BEGIN
   select path into edges from path_log where path_log.name = path_name;
   return  energy_on_path_DT_interval(edges, start_time, factor1, factor2, max_energy);   
END; 
$$
LANGUAGE plpgsql;
-- test
 select s->'first'->>'time',s->'mid1'->>'time',s->'mid2'->>'time', s->'last'->>'time', 
        s->'first'->>'energy',s->'mid1'->>'energy', s->'mid2'->>'energy', s->'last'->>'energy',   
       s->'first'->>'path_time',s->'mid'->>'path_time',s->'mid'->>'path_speed',s->'last'->>'path_time', s
	   from energy_on_path_nameid_DT_interval('1-6', 16*36000, 0.25, 0.75 ) as s;


-- interval is defined by lowering speed and increasing speed from average speed independantly of the time of the day
-- plusfactor - means user will go + <plusfactor> percent faster. Values should be 1.xx  (> 1)
-- minusfactor - means user will go - <minusfactor> percent slower. Values should be 0.xx (< 1)

CREATE OR REPLACE FUNCTION energy_on_path_speedy_interval(edges text[], start_time int, minusfactor float, plusfactor float,  factor1 float, factor2 float, max_energy float default null, car_physics json default null) RETURNS json AS $$
DECLARE
  int_first    json;
  int_last     json;
  
  res          json;  
  
BEGIN
  
  int_first := energy_on_path(edges, start_time, 0, 0, minusfactor, max_energy, car_physics);
  int_last  := energy_on_path(edges, start_time, 0, 0, plusfactor, max_energy, car_physics);
  
  IF (int_first->>'time')::FLOAT > (int_last->>'time')::FLOAT  THEN
    res := energy_on_path_interval_clipper(int_last, int_first, edges, start_time, factor1, factor2, max_energy, car_physics);
  ELSE
    res := energy_on_path_interval_clipper(int_first, int_last, edges, start_time, factor1, factor2, max_energy, car_physics);
  END IF;
  
  RETURN res;
END;
$$ LANGUAGE plpgsql; 


-- select s->'first'->>'time',s->'mid1'->>'time',s->'mid2'->>'time', s->'last'->>'time', 
--        s->'first'->>'energy',s->'mid1'->>'energy', s->'mid2'->>'energy', s->'last'->>'energy',   
--        s->'first'->>'path_time',s->'mid'->>'path_time',s->'mid'->>'path_speed',s->'last'->>'path_time', s
--	   from energy_on_path_speedy_interval('{"1-26731","26731-26735","26735-363051","363051-119619","119619-243307","243307-404182","404182-404183","404183-135287","135287-403967","403967-403968","403968-403980","403980-235853"}'::text[], 16*36000, 0.8, 1.05, 0.25, 0.75 ) as s;


-- convinience function will calculte energy by given path_id which is a id in table path_log.
CREATE OR REPLACE FUNCTION energy_on_path_id_speedy_interval(path_id integer, start_time int, minusfactor float, plusfactor float,  factor1 float, factor2 float, max_energy float default null, car_physics json default null) RETURNS json AS $$
DECLARE
   edges text[];
BEGIN
   select path into edges from path_log where path_log.id = path_id;
   return  energy_on_path_speedy_interval(edges, start_time, minusfactor, plusfactor,  factor1, factor2, max_energy, car_physics);   
END; 
$$
LANGUAGE plpgsql;
-- test
 select s->'first'->>'time',s->'mid1'->>'time',s->'mid2'->>'time', s->'last'->>'time', 
        s->'first'->>'energy',s->'mid1'->>'energy', s->'mid2'->>'energy', s->'last'->>'energy',   
        s->'first'->>'path_time',s->'mid'->>'path_time',s->'mid'->>'path_speed',s->'last'->>'path_time', s
	   from energy_on_path_id_speedy_interval(1, 16*36000, 0.8, 1.05, 0.25, 0.75 ) as s;

-- convinience function will calculte energy by given path_namer which is a name in table path_log.
CREATE OR REPLACE FUNCTION energy_on_path_nameid_speedy_interval(path_name text, start_time int, minusfactor float, plusfactor float,  factor1 float, factor2 float, max_energy float default null, car_physics json default null) RETURNS json AS $$
DECLARE
   edges text[];
BEGIN
   select path into edges from path_log where path_log.name = path_name;
   return  energy_on_path_speedy_interval(edges, start_time, minusfactor, plusfactor,  factor1, factor2, max_energy, car_physics);   
END; 
$$
LANGUAGE plpgsql;
-- test
 select s->'first'->>'time',s->'mid1'->>'time',s->'mid2'->>'time', s->'last'->>'time', 
        s->'first'->>'energy',s->'mid1'->>'energy', s->'mid2'->>'energy', s->'last'->>'energy',   
        s->'first'->>'path_time',s->'mid'->>'path_time',s->'mid'->>'path_speed',s->'last'->>'path_time', s
	   from energy_on_path_nameid_speedy_interval('1-6', 16*36000, 0.8, 1.05, 0.25, 0.75 ) as s;



-- ttl_interval_bounds will calculate time slot bounds for given time (it will do mod 24 hour, so 25 will be converted to 24)

CREATE OR REPLACE FUNCTION ttl_interval_bounds(start_time int, int_size int default 9000) RETURNS int[] AS $$
DECLARE
  time_first  int := 0;
  time_last   int := 0;
  slot        int := 0;
BEGIN
    
 -- time_first := mod(864000 + int_size*floor(start_time/int_size)::int, 864000);
 -- time_last  := mod(int_size*(floor(start_time/int_size)::int + 1), 864000);
 time_first := mod(864000 + start_time - 36000*2, 864000);
 time_last  := mod(start_time + 36000*2, 864000);
 
 select time, price from trafic_profile where time > time_first and time < time_last
 
  slot       := floor(start_time/int_size)::int;
  
  RETURN array [time_first, time_last, slot];
END;
$$ LANGUAGE plpgsql;

-- test
--select ttl_interval_bounds(700000);



CREATE OR REPLACE FUNCTION energy_on_path_interval(edges int[], start_time int, granularity int default 0 ) 
RETURNS  table (ttm float, energy float) AS $$
DECLARE
 
  time_first  int := 0;
  time_last   int := 0;

  int_first   float[];
  int_last    float[];
  int_tmp     float[];
  scale       float;
  
  int_size    int := 0;   

  cnt         int := 1;
  
BEGIN
 
  time_first := mod(864000 + 36000*floor((start_time - 36000)/36000)::int, 864000);
  time_last  := mod(36000*floor((start_time + 36000)/36000)::int, 864000);

  RAISE NOTICE 'time first % time last %', time_first, time_last;

  int_first := energy_on_path(edges, time_first, 0);
  int_last  := energy_on_path(edges, time_last, 0);
  
  RAISE NOTICE 'int first % int last %', int_first, int_last;
  --int first {828038,21.5886940811043} int last {36038,21.6986066266257}
  
  int_size  := abs(int_last[1]*10 - int_first[1]*10)/granularity;
  scale     := sign(int_last[1] - int_first[1])*granularity/(int_first[1]*10);
  
  RAISE NOTICE ' size % > scale % , granularity % ', int_size, scale, granularity;
  
  ttm := int_first[1]*10; -- KATCH uses 0.1 sec
  energy := int_first[2];
  return next;  

  LOOP
    EXIT WHEN cnt >= int_size - 1;
	int_tmp := energy_on_path(edges, time_first, scale*cnt);
	ttm := int_tmp[1]*10; -- KATCH uses 0.1 sec
	energy := int_tmp[2];
	return next;

   cnt := cnt + 1;
  END LOOP;
  
  ttm := int_last[1]*10; -- KATCH uses 0.1 sec
  energy := int_last[2];
  return next;

END;
$$ LANGUAGE plpgsql;

select * from energy_on_path_interval('{32,33,44,55,66,77,88,99}'::INT[], 16*36000, 1000 );

select * from energy_on_path('{32,33,44,55,66,77,88,99}'::INT[], 16*36000, 0, 100 );


select energy_on_path('{32,33}'::INT[], 36000*a), a from generate_series(0, 23) as a;

select energy_on_path('{32,33}'::INT[], 36000, 0.1*a), a from generate_series(0, 5) as a;

select energy_on_path('{32,33}'::INT[], 0, 0.1*a), a from generate_series(0, 5) as a;

select energy_on_path_interval('{32,33}'::INT[], 0, granularity int default  );






CREATE OR REPLACE FUNCTION energy_on_path_time_slot(edges int[], start_time int, interval_size int default 36000 ) 
RETURNS  table (ttm float, energy float) AS $$
DECLARE
 
  time_first  int := 0;
  time_last   int := 0;
  time_tmp    int := 0;

  slot_start  int[];
  slot_first  int[];
  slot_last   int[]; 
  
  int_first   float[];
  int_last    float[];
  int_tmp     float[];
  scale       float;
  
  direction   int; -- +1 positive, -1 negative
  
  int_size    int := 0;   

  cnt         int := 1;
  
BEGIN
  
  slot_start := ttl_interval_bounds(start_time, interval_size);

  RAISE NOTICE 'time first % time last %', slot_start[1], slot_start[2];

  int_first := energy_on_path(edges, slot_start[1], 0);
  int_last  := energy_on_path(edges, slot_start[2], 0);
  
  RAISE NOTICE 'int first % int last %', int_first, int_last;
  --int first {828038,21.5886940811043} int last {36038,21.6986066266257}
  
  slot_first := ttl_interval_bounds(slot_start[1] + int_first[1], interval_size);
  slot_last  := ttl_interval_bounds(slot_start[2] + int_last[1], interval_size);
  direction := sign(int_last[1] - int_first[1]);
  --int_size  := abs(int_last[1]*10 - int_first[1]*10)/interval_size;
  
  --RAISE NOTICE ' size % > scale % , granularity % ', int_size, scale, granularity;
  
  ttm := int_first[1]*10; -- KATCH uses 0.1 sec
  energy := int_first[2];
  return next;  

  time_tmp := slot_first[(1.5 + 0.5*direction)::int]; -- map 1 to 2;  -1 to 1
  
  LOOP
    EXIT WHEN ((direction > 0) AND time_tmp > slot_last[1]) OR 
	          ((direction < 0) AND time_tmp < slot_last[2]);
	
	scale     := (time_tmp - int_first[1]*10)/(int_first[1]*10);
	
	int_tmp := energy_on_path(edges, time_first, scale*cnt);
	ttm := int_tmp[1]*10; -- KATCH uses 0.1 sec
	energy := int_tmp[2];
	return next;

   cnt := cnt + 1;
  END LOOP;
  
  ttm := int_last[1]*10; -- KATCH uses 0.1 sec
  energy := int_last[2];
  return next;

END;
$$ LANGUAGE plpgsql;