-- https://gis.stackexchange.com/questions/334020/using-st-segmentize-to-postgresql-table-of-segments
--------------------------------------------------------------------------------------
--------------------------------------------------------------------------------------
-- dl_elevations will return an elevation profile for a given polyline geometry object
-- Params: geometry is a polyline object, integer is the maximum distance between profiling elevation points
DROP FUNCTION dl_elevations(geometry,integer);	
CREATE OR REPLACE FUNCTION dl_elevations(geometry, step int DEFAULT 50) 
                  RETURNS TABLE(id bigint, geom geometry, elev double precision)  AS $$ 
  WITH
    densified AS (
      SELECT ST_Segmentize($1::GEOGRAPHY, $2) as geom -- every 50m
    )
  SELECT 
    ROW_NUMBER() OVER() AS id,
    dmp.geom, --dmp.path,
    ST_Value(dem.rast, dmp.geom) AS ele 
  FROM 
    densified as l,	  
    LATERAL ST_DumpPoints(l.geom::GEOMETRY) AS dmp,	  
    public.de_srtm dem
  WHERE
    ST_Intersects(dmp.geom, dem.rast)
  ORDER BY id ASC
$$
LANGUAGE SQL;
--------------------------------------------------------------------------------------
-- json version of dl_elevations (gievn geometry as geojson, return points as geojson)
DROP FUNCTION dl_elevations(text,integer);	
CREATE OR REPLACE FUNCTION dl_elevations(geomjson text, step int DEFAULT 50) 
                  RETURNS TABLE(id bigint, geom text, elev double precision)  AS $$ 
	SELECT id, ST_AsGeoJSON(geom), elev from dl_elevations(ST_GeomFromGeoJSON($1),$2)
$$
LANGUAGE SQL;



-------------------------------------------------------------------------------------- 
--dl_elevations_mapped function augments the elevation profile with distances between two points
--------------------------------------------------------------------------------------
DROP FUNCTION dl_elevations_mapped(float8, geometry, float8, float8);	
CREATE OR REPLACE FUNCTION dl_elevations_mapped(geometry, step int DEFAULT 50)
                  RETURNS TABLE(elev float8, geom geometry, delta float8, delta_dist float8)  AS $$ 
	WITH profile AS (
	  select id, elev, geom from dl_elevations($1, $2)
	)
	select 
	  a.elev, a.geom, b.elev - a.elev, ST_Distance(a.geom::GEOGRAPHY, b.geom::GEOGRAPHY)	  
	  from profile a  LEFT JOIN  profile b ON a.id + 1= b.id 
	  ORDER by a.id;
$$
LANGUAGE SQL;
--------------------------------------------------------------------------------------
--DROP FUNCTION dl_elevations_mapped(text, int);	
CREATE OR REPLACE FUNCTION dl_elevations_mapped(text, step int DEFAULT 50)
                  RETURNS TABLE(elev float8, geojson text, delta float8, delta_dist float8)  AS $$ 
	select  elev, ST_AsGeoJSON(geom), delta, delta_dist from dl_elevations_mapped(ST_GeomFromGeoJSON($1), $2) 
$$
LANGUAGE SQL;
--------------------------------------------------------------------------------------

-------------------------------------------------------------------------------------- 
--------------------------------------------------------------------------------------
DROP FUNCTION dl_lineclimb(geometry, integer, bool);	
CREATE OR REPLACE FUNCTION dl_lineclimb(geometry, step int DEFAULT 50, distances bool DEFAULT TRUE)
                  RETURNS TABLE(climb float8, climb_dist float8, roll float8, roll_dist float8)  AS $$ 
	WITH profile AS (
	  select id, elev, geom from dl_elevations($1, $2)
	)
	select 
	  SUM (CASE WHEN b.elev > a.elev THEN b.elev - a.elev
	                                 ELSE 0 END
	       ) AS climb,
	  SUM (CASE WHEN $3 AND (b.elev > a.elev) THEN ST_Distance(a.geom::GEOGRAPHY, b.geom::GEOGRAPHY)
	                                 ELSE 0 END
	       ) AS climb_dist,
	  SUM (CASE WHEN b.elev < a.elev THEN a.elev - b.elev
	                                 ELSE 0 END
	      ) AS roll,
	  SUM (CASE WHEN $3 AND (b.elev < a.elev) THEN ST_Distance(a.geom::GEOGRAPHY, b.geom::GEOGRAPHY)
	                                 ELSE 0 END
	      ) AS roll_dist
	  from profile a  JOIN  profile b ON a.id + 1 = b.id;
$$
LANGUAGE SQL;

--------------------------------------------------------------------------------------
DROP FUNCTION dl_lineclimb(text, integer, bool);	
CREATE OR REPLACE FUNCTION dl_lineclimb(text, step int DEFAULT 50, distances bool DEFAULT TRUE)
                  RETURNS TABLE(climb float8, climb_dist float8, roll float8, roll_dist float8)  AS $$ 
	select  climb, climb_dist, roll, roll_dist from dl_lineclimb(ST_GeomFromGeoJSON($1), $2, $3) 
$$
LANGUAGE SQL;
--------------------------------------------------------------------------------------


-- tests
WITH
  polyline AS (
    SELECT pk.geom
	FROM public.pagrindiniai_keliai pk
    where pk.id = 340400
  )	
select elpr.* from polyline pl, LATERAL dl_elevations(pl.geom,200) elpr;

WITH
  polyline AS (
    SELECT pk.geom
	FROM public.pagrindiniai_keliai pk
    where pk.id = 340400
  )	
select elpr.* from polyline pl, LATERAL dl_elevations(ST_AsGeoJSON(pl.geom),200) elpr;

select id, elev, geom 
from dl_elevations(ST_GeomFromGeoJSON('{"type": "LineString","coordinates": [[7.65129,47.88411],[7.841491,47.88181]]}'),
				   30);

select id, elev, geom 
from dl_elevations('{"type": "LineString","coordinates": [[7.65129,47.88411],[7.841491,47.88181]]}',
				   30);
				   

select *
from dl_elevations_mapped(ST_GeomFromGeoJSON('{"type": "LineString","coordinates": [[7.65129,47.88411],[7.841491,47.88181]]}'),
				   300);

select *
from dl_elevations_mapped('{"type": "LineString","coordinates": [[7.65129,47.88411],[7.841491,47.88181]]}',
				   300);



select * from dl_lineclimb(ST_GeomFromGeoJSON('{"type": "LineString","coordinates": [[7.65129,47.88411],[7.841491,47.88181]]}'),30);

select * from dl_lineclimb(ST_GeomFromGeoJSON('{"type": "LineString","coordinates": [[7.65129,47.88411],[7.841491,47.88181]]}'),30, false);


select * from dl_lineclimb('{"type": "LineString","coordinates": [[7.65129,47.88411],[7.841491,47.88181]]}',30, true);
